import functools
import logging
import random
import shutil
import tempfile
from pathlib import Path
from typing import Any, Callable, List, Union, cast

import numpy as np

from gobbli.augment.base import BaseAugment
from gobbli.util import (
    TokenizeMethod,
    assert_in,
    assert_type,
    download_archive,
    download_file,
    is_archive,
    tokenize,
)

LOGGER = logging.getLogger(__name__)

_WORD2VEC_MODEL_ARCHIVES = {
    "glove.6B": "http://nlp.stanford.edu/data/glove.6B.zip",
    "glove.42B.300d": "http://nlp.stanford.edu/data/glove.42B.300d.zip",
    "glove.840B.300d": "http://nlp.stanford.edu/data/glove.840B.300d.zip",
    "glove.twitter.27B": "http://nlp.stanford.edu/data/glove.twitter.27B.zip",
    "fasttext.en.300d": "https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/wiki.en.vec",
    "fasttext.simple.300d": "https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/wiki.simple.vec",
    "charngram.100d": "http://www.logos.t.u-tokyo.ac.jp/~hassy/publications/arxiv2016jmt/jmt_pre-trained_embeddings.tar.gz",
}
"""
A mapping from archive names to downloadable archives.  See :obj:`WORD2VEC_MODELS`
for a list of names users can pass to get pretrained word2vec models.
"""

WORD2VEC_MODELS = {
    "glove.6B.50d": ("glove.6B", "glove.6B.50d.txt"),
    "glove.6B.100d": ("glove.6B", "glove.6B.100d.txt"),
    "glove.6B.200d": ("glove.6B", "glove.6B.200d.txt"),
    "glove.6B.300d": ("glove.6B", "glove.6B.300d.txt"),
    "glove.42B.300d": ("glove.42B.300d", "glove.42B.300d.txt"),
    "glove.840B.300d": ("glove.840B.300d", "glove.840B.300d.txt"),
    "glove.twitter.27B.25d": ("glove.twitter.27B", "glove.twitter.27B.25d"),
    "glove.twitter.27B.50d": ("glove.twitter.27B", "glove.twitter.27B.50d"),
    "glove.twitter.27B.100d": ("glove.twitter.27B", "glove.twitter.27B.100d"),
    "glove.twitter.27B.200d": ("glove.twitter.27B", "glove.twitter.27B.200d"),
    "fasttext-en": ("fasttext.en.300d", "wiki.en.vec"),
    "fasttext-simple": ("fasttext.simple.300d", "wiki.simple.vec"),
    "charngram": ("charngram.100d", "charNgram.txt"),
}
"""
A mapping from word2vec model names to archive URLs and filenames (since
some models contain multiple files per archive).  Pass one of these key names
to :class:`Word2Vec` to use pretrained model weights.
"""


class Word2Vec(BaseAugment):
    """
    Data augmentation method based on word2vec.  Replaces words with similar words
    according to vector similarity.

    Args:
      model: Pretrained word2vec model to use.  If a string, it should correspond
        to one of the keys in :obj:`WORD2VEC_MODELS`.  The corresponding pretrained
        vectors will be downloaded and used.  If a Path, it's assumed to be
        a file containing pretrained vectors, which will be loaded into a gensim word2vec
        model.  If a gensim Word2Vec model, it will be used directly.
      tokenizer: Function used for tokenizing texts to do word replacement. If an instance
        of :class:`gobbli.util.TokenizeMethod`, the corresponding preset tokenization method
        will be used.  If a callable, it should accept a list of strings and return a
        list of tokenized strings.
      n_similar: Number of similar words to be considered for replacement.
      diversity: 0 < diversity <= 1; determines the likelihood of selecting replacement words
        based on their similarity to the original word.
        At 1, the most similar words are most likely to be selected
        as replacements.  As diversity decreases, likelihood of selection becomes less
        dependent on similarity.
    """

    def __init__(
        self,
        # Can't make this type more restrictive since gensim might not be
        # available, and we need to make the union include a gensim type
        model: Any,
        tokenizer: Union[
            TokenizeMethod, Callable[[List[str]], List[List[str]]]
        ] = TokenizeMethod.SPLIT,
        n_similar: int = 10,
        diversity: float = 0.8,
    ):
        try:
            import gensim
            from gensim.scripts.glove2word2vec import glove2word2vec
        except ImportError:
            raise ImportError(
                "word2vec-based data augmentation requires gensim to be installed."
            )

        if isinstance(model, str):
            # Download and extract pretrained weights from a public source
            assert_in("word2vec model", model, set(WORD2VEC_MODELS.keys()))
            archive_name, filename = WORD2VEC_MODELS[model]
            archive_url = _WORD2VEC_MODEL_ARCHIVES[archive_name]

            LOGGER.debug(f"Downloading word2vec model '{model}'")
            # Some downloads aren't contained in archives
            if is_archive(Path(archive_url)):
                extract_dir = download_archive(
                    archive_url, self.data_dir(), junk_paths=True
                )
                model_file = extract_dir / filename
            else:
                model_file = download_file(archive_url)

            if model.startswith("glove"):
                LOGGER.debug(f"Converting GloVe format to word2vec format")
                # Need to convert the downloaded file to word2vec format,
                # since GloVe vectorsr are formatted slightly differently
                with tempfile.NamedTemporaryFile() as f:
                    tempfile_path = Path(f.name)
                    glove2word2vec(model_file, tempfile_path)
                    shutil.copy2(tempfile_path, model_file)

            LOGGER.debug(f"Loading word2vec model '{model}'")
            self._model = gensim.models.KeyedVectors.load_word2vec_format(model_file)
            LOGGER.debug(f"word2vec model loaded")
        elif isinstance(model, Path):
            LOGGER.debug(f"Loading word2vec model from path '{model}'")
            self._model = gensim.models.KeyedVectors.load_word2vec_format(str(model))
            LOGGER.debug(f"word2vec model loaded")
        elif isinstance(model, (gensim.models.Word2Vec, gensim.models.KeyedVectors)):
            self._model = model
        else:
            raise TypeError(
                f"unsupported type for initializing word2vec model: '{type(model)}'"
            )

        assert_type("n_similar", n_similar, int)
        if n_similar <= 0:
            raise ValueError("n_similar must be > 0")
        self.n_similar = n_similar

        assert_type("diversity", diversity, float)
        if not 0 < diversity <= 1:
            raise ValueError("diversity must be > 0 and <= 1")
        self.diversity = diversity

        if isinstance(tokenizer, TokenizeMethod):
            # Avoid mypy error when passing a partially-applied function created by
            # functools.partial
            self.tokenizer = cast(
                Callable[[List[str]], List[List[str]]],
                functools.partial(tokenize, tokenizer),
            )
        elif callable(tokenizer):
            self.tokenizer = tokenizer
        else:
            raise TypeError(f"unsupported type for tokenizer: '{type(tokenizer)}'")

    def _maybe_replace_token(self, token: str) -> str:
        """
        Generate a replacement for the given token using our model and parameters.
        """
        try:
            similar = self._model.most_similar(positive=token, topn=self.n_similar)
        except KeyError:
            # This token doesn't exist in the word2vec model's vocabulary --
            # don't change it
            return token
        similar_probs = np.array([s[1] for s in similar])

        # Scale according to diversity and normalize
        scaled_probs = np.power(similar_probs, 1 / self.diversity)
        scaled_probs = scaled_probs / np.sum(scaled_probs)
        selected_ndx = np.argmax(np.random.multinomial(1, scaled_probs))
        return similar[selected_ndx][0]

    def augment(self, X: List[str], times: int = 5, p: float = 0.1) -> List[str]:
        new_texts = []
        tokenized_texts = self.tokenizer(X)
        for _ in range(times):
            for tokens in tokenized_texts:
                new_tokens = tokens[:]
                for i in range(len(new_tokens)):
                    if random.random() < p:
                        new_tokens[i] = self._maybe_replace_token(new_tokens[i])
                # TODO can we do better with detokenization here?
                new_texts.append(" ".join(new_tokens))
        return new_texts



from dataclasses import dataclass
from pathlib import Path

DOCKER_ROOT = Path("/gobbli")


@dataclass
class ContainerTaskContext:
    """
    Encapsulates filesystem organization for tasks which take some
    input and produce some output using directories mapped between
    the host and a container.

    Provide necessary input/output directories but also allow the caller to create
    their own directories as needed and map them to container paths.

    Args:
      task_root_dir: The root directory for the task on the host.  All directories in the context
        will be descendents of this directory.
    """

    task_root_dir: Path

    def host_dir(self, name: str) -> Path:
        """
        Create (if necessary) and return a directory on the host under the task root path.

        Args:
          name: The name of the directory to create (relative to
            :paramref:`ContainerTaskContext.params.task_root_dir`).

        Returns:
          The full path to the created task root directory.
        """
        named_dir = self.task_root_dir / name
        named_dir.mkdir(parents=True, exist_ok=True)
        return named_dir

    def to_container(self, host_dir: Path) -> Path:
        """
        Convert a given directory on the host to a directory under some canonical
        path in the container.  The given host directory must be under our root
        directory.

        Args:
          host_dir: The full path to a descendent directory of the
            :paramref:`ContainerTaskContext.params.task_root_dir`).
        """
        return DOCKER_ROOT / host_dir.resolve().relative_to(
            self.task_root_dir.resolve()
        )

    @property
    def container_root_dir(self) -> Path:
        """
        Returns:
          The container root directory corresponding to the
          :paramref:`ContainerTaskContext.params.task_root_dir` host directory.
        """
        return self.to_container(self.task_root_dir)

    @property
    def host_input_dir(self) -> Path:
        """
        Returns:
          The host directory to be used for task input.
        """
        return self.host_dir("input")

    @property
    def host_output_dir(self) -> Path:
        """
        Returns:
          The host directory to be used for task output.
        """
        return self.host_dir("output")

    @property
    def container_input_dir(self) -> Path:
        """
        Returns:
          The directory to be used for task input, as mapped in the container.
        """
        return self.to_container(self.host_input_dir)

    @property
    def container_output_dir(self) -> Path:
        """
        Returns:
          The directory to be used for task output, as mapped in the container.
        """
        return self.to_container(self.host_output_dir)



from pathlib import Path
from typing import Any, Dict, List

from gobbli.augment.base import BaseAugment
from gobbli.docker import maybe_mount, run_container
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.util import assert_type, escape_line_delimited_texts


class BERTMaskedLM(BaseModel, BaseAugment):
    """
    BERT-based data augmenter.  Applies masked language modeling to generate
    predictions for missing tokens using a trained BERT model.
    """

    _BUILD_PATH = Path(__file__).parent

    _INPUT_FILE = "input.txt"
    _OUTPUT_FILE = "output.txt"

    _CONTAINER_CACHE_DIR = Path("/cache")

    def init(self, params: Dict[str, Any]):
        """
        See :meth:`gobbli.model.base.BaseModel.init`.

        BERTMaskedLM parameters:

        - ``bert_model`` (:obj:`str`): Name of a pretrained BERT model to use.
          See the `pytorch-transformers <https://huggingface.co/pytorch-transformers/pretrained_models.html>`__
          docs for supported values.
        - ``diversity``: 0 < diversity <= 1; determines the likelihood of selecting replacement words
          based on their predicted probability.
          At 1, the most probable words are most likely to be selected
          as replacements.  As diversity decreases, likelihood of selection becomes less
          dependent on predicted probability.
        - ``n_probable``: The number of probable tokens to consider for replacement.
        - ``batch_size``: Number of documents to run through the BERT model at once.
        """
        self.bert_model = "bert-base-uncased"
        self.diversity = 0.8
        self.batch_size = 32
        self.n_probable = 5

        for name, value in params.items():
            if name == "bert_model":
                self.bert_model = value
            elif name == "diversity":
                assert_type(name, value, float)
                if not 0 < value <= 1:
                    raise ValueError("diversity must be > 0 and <= 1")
                self.diversity = value
            elif name == "batch_size":
                assert_type(name, value, int)
                if value < 1:
                    raise ValueError("batch_size must be >= 1")
                self.batch_size = value
            elif name == "n_probable":
                assert_type(name, value, int)
                if value < 1:
                    raise ValueError("n_probable must be >= 1")
                self.n_probable = value
            else:
                raise ValueError(f"Unknown param '{name}'")

    @property
    def image_tag(self) -> str:
        """
        Returns:
          The Docker image tag to be used for the BERT container.
        """
        return f"gobbli-bert-maskedlm"

    def _build(self):
        self.docker_client.images.build(
            path=str(BERTMaskedLM._BUILD_PATH),
            tag=self.image_tag,
            **self._base_docker_build_kwargs,
        )

    def _write_input(self, X: List[str], context: ContainerTaskContext):
        """
        Write the user input to a file for the container to read.
        """
        input_path = context.host_input_dir / BERTMaskedLM._INPUT_FILE
        input_path.write_text(escape_line_delimited_texts(X))

    def _read_output(self, context: ContainerTaskContext) -> List[str]:
        """
        Read generated text output to a file from the container.
        """
        output_path = context.host_output_dir / BERTMaskedLM._OUTPUT_FILE
        return output_path.read_text().split("\n")

    @property
    def host_cache_dir(self):
        """
        Directory to be used for downloaded pytorch-transformers files.
        Should be the same across all instances of the class, since these are
        generally static model weights/config files that can be reused.
        """
        cache_dir = BERTMaskedLM.model_class_dir() / "cache"
        cache_dir.mkdir(exist_ok=True, parents=True)
        return cache_dir

    def augment(self, X: List[str], times: int = 5, p: float = 0.1) -> List[str]:
        context = ContainerTaskContext(self.data_dir())

        self._write_input(X, context)

        # Determine which device to use for augmentation
        device = "cpu"
        if self.use_gpu:
            if self.nvidia_visible_devices == "all":
                device = "cuda"
            else:
                device_num = self.nvidia_visible_devices.split(",")[0]
                device = f"cuda:{device_num}"

        cmd = (
            "python3 augment_text.py"
            f" {context.container_input_dir / BERTMaskedLM._INPUT_FILE}"
            f" {context.container_output_dir / BERTMaskedLM._OUTPUT_FILE}"
            f" --probability {p}"
            f" --times {times}"
            f" --diversity {self.diversity}"
            f" --bert-model {self.bert_model}"
            f" --batch-size {self.batch_size}"
            f" --n-probable {self.n_probable}"
            f" --cache-dir {BERTMaskedLM._CONTAINER_CACHE_DIR}"
            f" --device {device}"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        maybe_mount(
            run_kwargs["volumes"],
            self.host_cache_dir,
            BERTMaskedLM._CONTAINER_CACHE_DIR,
        )

        run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        return self._read_output(context)



import logging
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from timeit import default_timer as timer
from typing import Any, Dict, Optional

import docker

from gobbli.model.context import ContainerTaskContext
from gobbli.util import (
    format_duration,
    generate_uuid,
    gobbli_dir,
    is_dir_empty,
    read_metadata,
    write_metadata,
)

LOGGER = logging.getLogger(__name__)


def model_dir() -> Path:
    return gobbli_dir() / "model"


_WEIGHTS_DIR_NAME = "weights"


class BaseModel(ABC):
    """
    Abstract base class for all models.

    Derived classes should be careful to call super().__init__(...) with the appropriate
    arguments if they override __init__() to preserve all the functionality.

    Functionality to facilitate making GPU(s) available to derived classes is available.
    """

    _METADATA_FILENAME = "gobbli-model-meta.json"

    _WEIGHTS_DIR_NAME = _WEIGHTS_DIR_NAME
    _CONTAINER_WEIGHTS_PATH = Path("/model") / _WEIGHTS_DIR_NAME

    def __init__(
        self,
        data_dir: Optional[Path] = None,
        load_existing: bool = False,
        use_gpu: bool = False,
        nvidia_visible_devices: str = "all",
        logger: Optional[logging.Logger] = None,
        **kwargs,
    ):
        """
        Create a model.

        Args:
          data_dir: Optional path to a directory used to store model data.  If not given,
            a unique directory under GOBBLI_DIR will be created and used.
          load_existing: If True, ``data_dir`` should be a directory that was previously used
            to create a model.  Parameters will be loaded to match the original model, and
            user-specified model parameters will be ignored.  If False, the data_dir must
            be empty if it already exists.
          use_gpu: If True, use the
            nvidia-docker runtime (https://github.com/NVIDIA/nvidia-docker) to expose
            NVIDIA GPU(s) to the container.  Will cause an error if the computer you're running
            on doesn't have an NVIDIA GPU and/or doesn't have the nvidia-docker runtime installed.
          nvidia_visible_devices: Which GPUs to make available to the container; ignored if
            ``use_gpu`` is False.  If not 'all', should be a comma-separated string: ex. ``1,2``.
          logger: If passed, use this logger for logging instead of the default module-level logger.
          **kwargs: Additional model-specific parameters to be passed to the model's :meth:`init` method.
        """
        if data_dir is None:
            self._data_dir = self.model_class_dir() / generate_uuid()
        else:
            self._data_dir = data_dir
        self._data_dir.mkdir(parents=True, exist_ok=True)

        if load_existing and self.metadata_path.exists():
            params = read_metadata(self.metadata_path)
            if len(kwargs) > 0:
                warnings.warn(
                    "User-passed params ignored due to existing model being "
                    f"loaded: {kwargs}"
                )
        else:
            if not is_dir_empty(self._data_dir):
                raise ValueError(
                    f"data_dir '{self._data_dir}' is non-empty;"
                    " it must be empty to avoid overwriting data."
                )
            params = kwargs
            write_metadata(params, self.metadata_path)

        self.use_gpu = use_gpu
        self.nvidia_visible_devices = nvidia_visible_devices

        self._logger = LOGGER
        if logger is not None:
            self._logger = logger

        self.docker_client = docker.from_env()

        self.init(params)

    @property
    def logger(self) -> logging.Logger:
        """
        Returns:
          A logger for derived models to use.
        """
        return self._logger

    @property
    def metadata_path(self) -> Path:
        """
        Returns:
         The path to the model's metadata file containing model-specific parameters.
        """
        return self.data_dir() / BaseModel._METADATA_FILENAME

    @abstractmethod
    def init(self, params: Dict[str, Any]):
        """
        Initialize a derived model using parameters specific to that model.

        Args:
          params: A dictionary where keys are parameter names and values are
            parameter values.
        """
        raise NotImplementedError

    def _base_docker_run_kwargs(self, context: ContainerTaskContext) -> Dict[str, Any]:
        """
        Establish a base set of docker run kwargs to handle GPU support, etc.
        Map directories as specified by the context.

        Returns:
            Base kwargs for any model that will be run using Docker.
        """
        kwargs = {
            "environment": {
                # Minimize the probability of containers exiting without dumping
                # buffered output
                "PYTHONUNBUFFERED": "1"
            },
            "detach": True,
            "volumes": {
                str(context.task_root_dir): {
                    "bind": str(context.container_root_dir),
                    "mode": "rw",
                },
                # Ideally we'd mount this as read-only, but some models (e.g. fastText)
                # need to write to their weights
                str(self.weights_dir): {
                    "bind": str(BaseModel._CONTAINER_WEIGHTS_PATH),
                    "mode": "rw",
                },
            },
        }  # type: Dict[str, Any]

        if self.use_gpu:
            kwargs["environment"][
                "NVIDIA_VISIBLE_DEVICES"
            ] = self.nvidia_visible_devices
            kwargs["runtime"] = "nvidia"

        return kwargs

    @property
    def _base_docker_build_kwargs(self) -> Dict[str, Any]:
        """
        Handle GPU support, etc via common args for any model Docker container.

        Returns:
            Base kwargs for any model that will be built using Docker.
        """
        kwargs = {"buildargs": {}}  # type: Dict[str, Any]

        if self.use_gpu:
            kwargs["buildargs"]["GPU"] = "1"

        return kwargs

    def data_dir(self) -> Path:
        """
        Returns:
            The main data directory unique to this instance of the model.
        """
        return self._data_dir

    @classmethod
    def model_class_dir(cls) -> Path:
        """
        Returns:
          A directory shared among all classes of the model.
        """
        return model_dir() / cls.__name__

    @property
    def weights_dir(self) -> Path:
        """
        The directory used to store initial model weights (before fine-tuning).
        These should generally be some pretrained weights made available by model
        developers.  This directory will NOT be created by default; models should
        download their weights and remove the weights directory if the download doesn't
        finish properly.

        Returns:
            The path to the weights directory.
        """
        return self.data_dir() / BaseModel._WEIGHTS_DIR_NAME

    def build(self):
        """
        Perform any pre-setup that needs to be done before running the model
        (building Docker images, etc).
        """
        self.logger.info("Starting build.")
        start = timer()
        self._build()
        end = timer()
        self.logger.info(f"Build finished in {format_duration(end - start)}.")

    @abstractmethod
    def _build(self):
        """
        Used for derived classes to define their implementation of the build method.
        """
        raise NotImplementedError



import gzip
import io
import tarfile
import zipfile
from pathlib import Path
from typing import List

import pytest

from gobbli.util import (
    TokenizeMethod,
    blob_to_dir,
    detokenize,
    dir_to_blob,
    extract_archive,
    is_archive,
    tokenize,
)


def make_zip(tmpdir: Path, relative_paths: List[Path]) -> Path:
    """
    Make a zip archive from a list of relative paths.
    Create empty files at each path and add them to the archive.
    """
    zip_path = tmpdir / "test.zip"
    with zipfile.ZipFile(zip_path, "w") as z:
        for relative_path in relative_paths:
            full_path = tmpdir / relative_path
            full_path.parent.mkdir(exist_ok=True, parents=True)
            full_path.touch()
            z.write(full_path, arcname=relative_path)

    return zip_path


def make_tar_gz(tmpdir: Path, relative_paths: List[Path]) -> Path:
    """
    Make a .tar.gz archive from a list of relative paths.
    Create empty files at each path and add them to the archive.
    """
    tar_path = tmpdir / "test.tar.gz"
    with tarfile.open(tar_path, "w:gz") as z:
        for relative_path in relative_paths:
            full_path = tmpdir / relative_path
            full_path.parent.mkdir(exist_ok=True, parents=True)
            full_path.touch()
            z.add(str(full_path), arcname=str(relative_path), recursive=False)

    return tar_path


def make_gz(tmpdir: Path, name: str) -> Path:
    """
    Create a gzip-compressed file with the given name under the given temp directory.
    Return the path to the compressed file.
    """
    gzip_path = tmpdir / f"{name}.gz"
    with gzip.open(gzip_path, "wb") as z:
        z.write(b"Test")

    return gzip_path


TEST_ARCHIVE_DATA = ["./a", "./b/c"]


@pytest.mark.parametrize(
    "archive_func,junk,expected_paths",
    [
        (make_zip, False, [Path("a"), Path("b") / "c"]),
        (make_zip, True, [Path("a"), Path("c")]),
        (make_tar_gz, False, [Path("a"), Path("b") / "c"]),
        (make_tar_gz, True, [Path("a"), Path("c")]),
    ],
)
def test_extract_archive(tmpdir, archive_func, junk, expected_paths):
    tmpdir_path = Path(tmpdir)
    archive_path = archive_func(tmpdir_path, TEST_ARCHIVE_DATA)

    archive_extract_dir = tmpdir_path / "extract"
    extract_archive(archive_path, archive_extract_dir, junk_paths=junk)

    for relative_path in expected_paths:
        assert (archive_extract_dir / relative_path).exists()


def test_extract_gz(tmpdir):
    tmpdir_path = Path(tmpdir)
    filename = "test.txt"
    archive_path = make_gz(tmpdir_path, "test.txt")

    archive_extract_dir = tmpdir_path / "extract"
    extract_archive(archive_path, archive_extract_dir)

    assert (archive_extract_dir / filename).exists()


@pytest.mark.parametrize(
    "name,expected_is_archive",
    [
        ("test.tar.gz", True),
        ("test.gz", True),
        ("test.txt.gz", True),
        ("test.zip", True),
        ("test.xz", False),
        ("test.txt", False),
        ("test.vec", False),
        ("test.bin", False),
    ],
)
def test_is_archive(name, expected_is_archive):
    assert is_archive(Path(name)) == expected_is_archive


def test_dir_to_blob(tmpdir):
    test_dir = Path(tmpdir) / "test"
    test_dir.mkdir()
    test_file_name = "test.txt"
    test_file = test_dir / test_file_name
    file_contents = "test"
    test_file.write_text(file_contents)

    blob = dir_to_blob(test_dir)
    fileobj = io.BytesIO(blob)
    fileobj.seek(0)
    extract_path = test_dir / "test2"
    with tarfile.open(fileobj=fileobj, mode="r:gz") as archive:
        archive.extractall(extract_path)

    extracted_file = extract_path / test_file_name
    assert extracted_file.exists()
    assert extracted_file.read_text() == file_contents


def test_blob_to_dir(tmpdir):
    test_dir = Path(tmpdir) / "test"
    test_dir.mkdir()
    test_file_name = "test.txt"
    test_file = test_dir / test_file_name
    file_contents = "test"
    test_file.write_text(file_contents)

    blob = dir_to_blob(test_dir)
    extract_path = test_dir / "test2"
    blob_to_dir(blob, extract_path)

    extracted_file = extract_path / test_file_name
    assert extracted_file.exists()
    assert extracted_file.read_text() == file_contents


@pytest.mark.parametrize(
    "text,tokens",
    [
        ("This is a test.", ["this", "is", "a", "test."]),
        ("Two  spaces", ["two", "spaces"]),
        ("Hyphenated-word", ["hyphenated-word"]),
        ("Numbers 1 and 2", ["numbers", "1", "and", "2"]),
    ],
)
def test_tokenize_split(text, tokens):
    # Whitespace tokenization just splits on whitespace
    assert tokenize(TokenizeMethod.SPLIT, [text]) == [tokens]


@pytest.mark.parametrize(
    "text,tokens",
    [
        ("This is a test.", ["this", "is", "a", "test"]),
        ("Two  spaces", ["two", "spaces"]),
        ("Hyphenated-word", ["hyphenated", "word"]),
        ("Numbers 1 and 2", ["numbers", "and"]),
    ],
)
def test_tokenize_spacy(text, tokens):
    # Spacy tokenization lowercases and removes non-alphabetic tokens
    assert tokenize(TokenizeMethod.SPACY, [text]) == [tokens]


@pytest.mark.parametrize(
    "tokenize_method", [TokenizeMethod.SPLIT, TokenizeMethod.SPACY]
)
@pytest.mark.parametrize(
    "tokens,text",
    [
        (["this", "is", "a", "test"], "this is a test"),
        (["hyphenated-word"], "hyphenated-word"),
        (["try", ",", "punctuation", "."], "try , punctuation ."),
    ],
)
def test_detokenize_split_spacy(text, tokens, tokenize_method):
    assert detokenize(tokenize_method, [tokens]) == [text]


def test_tokenize_detokenize_sentencepiece(tmpdir):
    texts = ["a b c", "a ab c", "a b ac"]

    # Model should be trained
    model_path = Path(tmpdir) / "spm"
    tokens = tokenize(
        TokenizeMethod.SENTENCEPIECE, texts, model_path=model_path, vocab_size=7
    )

    # Control sequence indicating whitespace
    _ = "▁"
    expected_tokens = [
        [_, "a", _, "b", _, "c"],
        [_, "a", _, "a", "b", _, "c"],
        [_, "a", _, "b", _, "a", "c"],
    ]
    assert tokens == expected_tokens

    assert detokenize(TokenizeMethod.SENTENCEPIECE, tokens, model_path) == texts

    # Previously should be reused with the old vocab size, and a new model
    # shouldn't be trained
    tokens = tokenize(TokenizeMethod.SENTENCEPIECE, texts, model_path=model_path)
    assert tokens == expected_tokens



import gobbli.augment as augment
import gobbli.dataset as dataset
import gobbli.experiment as experiment
import gobbli.io as io
import gobbli.model as model
from gobbli.util import TokenizeMethod

__all__ = [
    # Modules
    "augment",
    "dataset",
    "experiment",
    "model",
    "io",
    # Misc top level imports
    "TokenizeMethod",
]



import json
import shutil
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

import gobbli.io
from gobbli.docker import maybe_mount, run_container
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import EmbedMixin, PredictMixin, TrainMixin
from gobbli.util import (
    assert_in,
    assert_type,
    download_archive,
    escape_line_delimited_texts,
)


def _preprocess_text(text_series: pd.Series) -> pd.Series:
    """
    Preprocess a Series of text for a BERT model.
    """
    return text_series.str.replace(r"\s", " ", regex=True)


def _df_to_train_tsv(df: pd.DataFrame, output_file: Path):
    """
    Write a dataframe with "Text" and "Label" columns to the given
    output file in the format expected by BERT for a TSV file for
    training/validation.
    """
    ordered_df = pd.DataFrame(
        OrderedDict(
            (
                ("Label", df["Label"]),
                ("a", np.repeat("a", df.shape[0])),
                ("Text", _preprocess_text(df["Text"])),
            )
        )
    )
    ordered_df.to_csv(output_file, sep="\t", index=True, header=False)


def _write_labels(labels: List[Any], output_file: Path):
    """
    Write the given set of labels to the given file.
    """
    output_file.write_text(escape_line_delimited_texts(labels))


def _read_predictions(labels: List[Any], output_file: Path) -> pd.DataFrame:
    """
    Read predictions from the BERT model into a dataframe containing the predicted
    probability for each label for each observation.
    """
    return pd.read_csv(output_file, sep="\t", names=labels)


def _read_embeddings(
    output_file: Path, pooling: gobbli.io.EmbedPooling
) -> Tuple[List[np.ndarray], Optional[List[List[str]]]]:
    """
    Read embeddings from the BERT model into a ndarray with the embedding values.  Also return
    a corresponding list of tokens if no pooling was applied.
    """
    embeddings = []  # type: List[np.ndarray]
    doc_tokens = []  # type: List[List[str]]
    with open(output_file, "r") as f:
        for line in f:
            line_json = json.loads(line)
            line_tokens = []  # type: List[str]
            line_layers = []  # type: List[List[float]]
            for token_info in line_json["features"]:
                line_tokens.append(token_info["token"])
                # Only take the first layer output; we don't currently support a
                # way to combine multiple layers
                line_layers.append(token_info["layers"][0]["values"])

            # Apply pooling, if necessary
            line_embedding = np.array(line_layers)
            if pooling == gobbli.io.EmbedPooling.MEAN:
                line_embedding = np.mean(line_embedding, axis=0)

            embeddings.append(line_embedding)
            doc_tokens.append(line_tokens)

    # Don't return tokens if we're doing any pooling, since
    # the pooled results combine all the tokens
    tokens = None
    if pooling == gobbli.io.EmbedPooling.NONE:
        tokens = doc_tokens

    return embeddings, tokens


BERT_MODEL_ARCHIVES = {
    "bert-base-uncased": "https://storage.googleapis.com/bert_models/2018_10_18/uncased_L-12_H-768_A-12.zip",
    "bert-base-cased": "https://storage.googleapis.com/bert_models/2018_10_18/cased_L-12_H-768_A-12.zip",
    "bert-large-uncased": "https://storage.googleapis.com/bert_models/2018_10_18/uncased_L-24_H-1024_A-16.zip",
    "bert-large-cased": "https://storage.googleapis.com/bert_models/2018_10_18/cased_L-24_H-1024_A-16.zip",
    "bert-large-whole-word-masking-uncased": "https://storage.googleapis.com/bert_models/2019_05_30/wwm_uncased_L-24_H-1024_A-16.zip",
    "bert-large-whole-word-masking-cased": "https://storage.googleapis.com/bert_models/2019_05_30/wwm_cased_L-24_H-1024_A-16.zip",
    "bert-base-multilingual-cased": "https://storage.googleapis.com/bert_models/2018_11_23/multi_cased_L-12_H-768_A-12.zip",
    "bert-base-chinese": "https://storage.googleapis.com/bert_models/2018_11_03/chinese_L-12_H-768_A-12.zip",
    "clinical-bert-cased": "https://www.dropbox.com/s/8armk04fu16algz/pretrained_bert_tf.tar.gz?dl=1",
    "biobert-cased": "https://github.com/naver/biobert-pretrained/releases/download/v1.1-pubmed/biobert_v1.1_pubmed.tar.gz",
    "scibert-uncased": "https://s3-us-west-2.amazonaws.com/ai2-s2-research/scibert/tensorflow_models/scibert_scivocab_uncased.tar.gz",
    "scibert-cased": "https://s3-us-west-2.amazonaws.com/ai2-s2-research/scibert/tensorflow_models/scibert_scivocab_cased.tar.gz",
}  # type: Dict[str, str]
"""
A mapping from model names to archives.
See `the BERT repo <https://github.com/google-research/bert>`__ for guidelines on when
to use which model.  "bert-base-uncased" is a safe default for most situations.
Larger models require more time and GPU memory to run.
"""


class BERT(BaseModel, TrainMixin, PredictMixin, EmbedMixin):
    """
    Classifier/embedding wrapper for Google Research's BERT:
    https://github.com/google-research/bert
    """

    _BUILD_PATH = Path(__file__).parent

    _TRAIN_INPUT_FILE = "train.tsv"
    _VALID_INPUT_FILE = "dev.tsv"
    _TEST_INPUT_FILE = "test.tsv"
    _LABELS_INPUT_FILE = "labels.tsv"

    _TEST_OUTPUT_FILE = "test_results.tsv"

    _EMBEDDING_INPUT_FILE = "input.txt"
    _EMBEDDING_OUTPUT_FILE = "embeddings.jsonl"

    def init(self, params: Dict[str, Any]):
        """
        See :meth:`gobbli.model.base.BaseModel.init`.

        BERT parameters:

        - ``max_seq_length`` (:obj:`int`): The maximum total input sequence length after
          WordPiece tokenization.  Sequences longer than this will be truncated,
          and sequences shorter than this will be padded.  Default: 128
        - ``bert_model`` (:obj:`str`): Name of a pretrained BERT model to use.
          See :obj:`BERT_MODEL_ARCHIVES` for a listing of available BERT models.
        """
        self.max_seq_length = 128
        self.bert_model = "bert-base-uncased"

        for name, value in params.items():
            if name == "max_seq_length":
                assert_type(name, value, int)
                self.max_seq_length = value
            elif name == "bert_model":
                assert_in(name, value, set(BERT_MODEL_ARCHIVES.keys()))
                self.bert_model = value
            else:
                raise ValueError(f"Unknown param '{name}'")

    @property
    def image_tag(self) -> str:
        """
        Returns:
          The Docker image tag to be used for the BERT container.
        """
        device = "gpu" if self.use_gpu else "cpu"
        return f"gobbli-bert-classifier-{device}"

    @property
    def do_lower_case(self) -> bool:
        """
        Returns:
          Whether the BERT tokenizer should lowercase its input.
        """
        return "uncased" in self.bert_model

    def _build(self):
        # Download data if we don't already have it
        if not self.weights_dir.exists():
            self.weights_dir.mkdir(parents=True)
            try:
                self.logger.info("Downloading pre-trained weights.")
                download_archive(
                    BERT_MODEL_ARCHIVES[self.bert_model],
                    self.weights_dir,
                    junk_paths=True,
                )
                self.logger.info("Weights downloaded.")
            except Exception:
                # Don't leave the weights directory in a partially downloaded state
                if self.weights_dir.exists():
                    shutil.rmtree(self.weights_dir)
                raise

        # Build the docker image
        self.docker_client.images.build(
            path=str(BERT._BUILD_PATH),
            tag=self.image_tag,
            **self._base_docker_build_kwargs,
        )

    @staticmethod
    def _get_checkpoint(
        user_checkpoint: Optional[Path], context: ContainerTaskContext
    ) -> Tuple[Optional[Path], Path, str]:
        """
        Determines the host checkpoint directory, container checkpoint directory, and
        checkpoint name using the user-requested checkpoint (if any) and the container context.

        Args:
          user_checkpoint: An optional checkpoint passed in by the user.  If the user doesn't
            pass one, use the default pretrained checkpoint.
          context: The container context to create the checkpoint in.

        Returns:
          A 3-tuple: the host checkpoint directory (if any),
            the container checkpoint directory, and the checkpoint name.
        """
        if user_checkpoint is None:
            # Default BERT weights
            host_checkpoint_dir = None
            checkpoint_name = "bert_model.ckpt"
            container_checkpoint_dir = BaseModel._CONTAINER_WEIGHTS_PATH
        else:
            # Trained weights, which will be mounted in the container
            host_checkpoint_dir = user_checkpoint.parent
            checkpoint_name = user_checkpoint.name
            container_checkpoint_dir = context.container_root_dir / "checkpoint"

        return host_checkpoint_dir, container_checkpoint_dir, checkpoint_name

    def _write_train_input(self, train_input: gobbli.io.TrainInput, input_dir: Path):
        """
        Write the given gobbli input into the format expected by BERT.
        Make sure the given directory exists first.
        """
        train_path = input_dir / BERT._TRAIN_INPUT_FILE
        valid_path = input_dir / BERT._VALID_INPUT_FILE

        train_df = pd.DataFrame(
            {"Text": train_input.X_train, "Label": train_input.y_train}
        )
        valid_df = pd.DataFrame(
            {"Text": train_input.X_valid, "Label": train_input.y_valid}
        )

        _df_to_train_tsv(train_df, train_path)
        _df_to_train_tsv(valid_df, valid_path)

        labels_path = input_dir / BERT._LABELS_INPUT_FILE
        _write_labels(train_input.labels(), labels_path)

    def _train(
        self, train_input: gobbli.io.TrainInput, context: ContainerTaskContext
    ) -> gobbli.io.TrainOutput:

        self._write_train_input(train_input, context.host_input_dir)

        # Determine checkpoint to use
        host_checkpoint_dir, container_checkpoint_dir, checkpoint_name = self._get_checkpoint(
            train_input.checkpoint, context
        )

        labels = train_input.labels()

        cmd = (
            "bash -c 'python run_classifier.py"
            " --task_name=cola"
            " --do_train=true"
            " --do_eval=true"
            f" --data_dir={context.container_input_dir}"
            f" --vocab_file={BaseModel._CONTAINER_WEIGHTS_PATH}/vocab.txt"
            f" --bert_config_file={BaseModel._CONTAINER_WEIGHTS_PATH}/bert_config.json"
            f" --init_checkpoint={BaseModel._CONTAINER_WEIGHTS_PATH}/bert_model.ckpt"
            f" --max_seq_length={self.max_seq_length}"
            f" --train_batch_size={train_input.train_batch_size}"
            f" --eval_batch_size={train_input.valid_batch_size}"
            f" --learning_rate=2e-5"
            f" --do_lower_case={self.do_lower_case}"
            f" --num_train_epochs={train_input.num_train_epochs}"
            f" --output_dir={context.container_output_dir}'"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the checkpoint in the container if needed
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint_dir, container_checkpoint_dir
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        # Parse the generated evaluation results file
        results_file = context.host_output_dir / "eval_results.txt"
        eval_results = {}  # type: Dict[str, Union[int, float]]
        with open(results_file, "r") as f:
            for line in f:
                key, str_val = line.split(" = ")
                if key == "global_step":
                    val: Union[int, float] = int(str_val)
                else:
                    val = float(str_val)
                eval_results[key] = val

        return gobbli.io.TrainOutput(
            valid_loss=eval_results["eval_loss"],
            valid_accuracy=eval_results["eval_accuracy"],
            train_loss=eval_results["loss"],
            labels=labels,
            checkpoint=context.host_output_dir
            / f"model.ckpt-{eval_results['global_step']}",
            _console_output=container_logs,
        )

    def _write_predict_input(
        self, predict_input: gobbli.io.PredictInput, input_dir: Path
    ):
        """
        Write the given gobbli prediction input into the format expected by BERT.
        Make sure the given directory exists first.
        """
        test_path = input_dir / BERT._TEST_INPUT_FILE

        test_df = pd.DataFrame({"sentence": predict_input.X})
        test_df["sentence"] = _preprocess_text(test_df["sentence"])
        test_df.index.name = "id"

        test_df.to_csv(test_path, sep="\t", index=True, header=True)

        labels_path = input_dir / BERT._LABELS_INPUT_FILE
        _write_labels(predict_input.labels, labels_path)

    def _predict(
        self, predict_input: gobbli.io.PredictInput, context: ContainerTaskContext
    ) -> gobbli.io.PredictOutput:

        self._write_predict_input(predict_input, context.host_input_dir)

        # Determine checkpoint to use
        host_checkpoint_dir, container_checkpoint_dir, checkpoint_name = self._get_checkpoint(
            predict_input.checkpoint, context
        )

        cmd = (
            "bash -c 'python run_classifier.py"
            " --task_name=cola"
            " --do_predict=true"
            f" --data_dir={context.container_input_dir}"
            f" --vocab_file={BaseModel._CONTAINER_WEIGHTS_PATH}/vocab.txt"
            f" --bert_config_file={BaseModel._CONTAINER_WEIGHTS_PATH}/bert_config.json"
            f" --predict-batch-size={predict_input.predict_batch_size}"
            f" --do_lower_case={self.do_lower_case}"
            f" --init_checkpoint={container_checkpoint_dir / checkpoint_name}"
            f" --max_seq_length={self.max_seq_length}"
            f" --output_dir={context.container_output_dir}'"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the checkpoint in the container if needed
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint_dir, container_checkpoint_dir
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        return gobbli.io.PredictOutput(
            y_pred_proba=_read_predictions(
                predict_input.labels, context.host_output_dir / BERT._TEST_OUTPUT_FILE
            ),
            _console_output=container_logs,
        )

    def _write_embed_input(self, embed_input: gobbli.io.EmbedInput, input_dir: Path):
        """
        Write the given gobbli embedding input into the format expected by BERT.
        Make sure the given directory exists first.
        """
        input_dir.mkdir(exist_ok=True, parents=True)
        input_path = input_dir / BERT._EMBEDDING_INPUT_FILE
        input_path.write_text(escape_line_delimited_texts(embed_input.X))

    def _embed(
        self, embed_input: gobbli.io.EmbedInput, context: ContainerTaskContext
    ) -> gobbli.io.EmbedOutput:

        self._write_embed_input(embed_input, context.host_input_dir)

        # Determine checkpoint to use
        host_checkpoint_dir, container_checkpoint_dir, checkpoint_name = self._get_checkpoint(
            embed_input.checkpoint, context
        )

        # Use the second-to-last layer for embeddings as suggested:
        # https://github.com/hanxiao/bert-as-service#q-why-not-the-last-hidden-layer-why-second-to-last
        cmd = (
            "bash -c 'python extract_features.py"
            f" --input_file={context.container_input_dir / BERT._EMBEDDING_INPUT_FILE}"
            f" --output_file={context.container_output_dir / BERT._EMBEDDING_OUTPUT_FILE}"
            f" --vocab_file={BaseModel._CONTAINER_WEIGHTS_PATH}/vocab.txt"
            f" --bert_config_file={BaseModel._CONTAINER_WEIGHTS_PATH}/bert_config.json"
            f" --init_checkpoint={container_checkpoint_dir / checkpoint_name}"
            f" --do_lower_case={self.do_lower_case}"
            f" --layers=-2"
            f" --max_seq_length={self.max_seq_length}"
            f" --batch_size={embed_input.embed_batch_size}'"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the checkpoint in the container if needed
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint_dir, container_checkpoint_dir
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        X_embedded, embed_tokens = _read_embeddings(
            context.host_output_dir / BERT._EMBEDDING_OUTPUT_FILE, embed_input.pooling
        )

        return gobbli.io.EmbedOutput(
            X_embedded=X_embedded,
            embed_tokens=embed_tokens,
            _console_output=container_logs,
        )



import numpy as np
import pandas as pd
import pytest

from gobbli.io import (
    EmbedOutput,
    PredictOutput,
    TrainOutput,
    WindowPooling,
    make_document_windows,
    pool_document_windows,
    validate_X,
    validate_X_y,
    validate_y,
)


@pytest.mark.parametrize("val_func", [validate_X, validate_y])
@pytest.mark.parametrize(
    "obj,err_class",
    [
        # Wrong collection type (series)
        (pd.Series("a"), TypeError),
        # Wrong collection type (numpy array)
        (np.array(["a"]), TypeError),
        # Wrong type (float)
        ([1.0], TypeError),
        # Wrong type (int)
        ([1], TypeError),
        # Correct type
        (["a"], None),
        # Empty
        ([], None),
    ],
)
def test_validate_input(obj, err_class, val_func):
    if err_class is None:
        val_func(obj)
    else:
        with pytest.raises(err_class):
            val_func(obj)


@pytest.mark.parametrize(
    "X,y,err_class",
    [
        # Same size, empty
        ([], [], None),
        # X is empty
        (["a"], [], ValueError),
        # y is empty
        ([], [1], ValueError),
        # X is longer
        (["a", "b"], [1], ValueError),
        # y is longer
        (["a"], [1, 2], ValueError),
        # Same size, non-empty
        (["a"], [1], None),
    ],
)
def test_validate_X_y(X, y, err_class):
    if err_class is None:
        validate_X_y(X, y)
    else:
        with pytest.raises(err_class):
            validate_X_y(X, y)


@pytest.mark.parametrize(
    "df,y_pred",
    [
        (pd.DataFrame({"a": [], "b": []}), []),
        (pd.DataFrame({"a": [1], "b": [0]}), ["a"]),
        (pd.DataFrame({"a": [0.7, 0.3], "b": [0.3, 0.7]}), ["a", "b"]),
    ],
)
def test_predict_output(df, y_pred):
    predict_output = PredictOutput(y_pred_proba=df)
    assert y_pred == predict_output.y_pred


@pytest.mark.parametrize("has_y", [True, False])
@pytest.mark.parametrize(
    "docs,window_len,expected_windowed",
    [
        # One doc, one window
        (["a"], 1, (["a"], [0])),
        # One doc, two windows
        (["a b"], 1, (["a", "b"], [0, 0])),
        # One doc, partial window
        (["a b c"], 2, (["a b", "c"], [0, 0])),
        # Two docs, one window
        (["a", "b"], 1, (["a", "b"], [0, 1])),
        # Two docs, two windows
        (["a b", "c d"], 1, (["a", "b", "c", "d"], [0, 0, 1, 1])),
        # Two docs, partial windows
        (["a b c", "d e f"], 2, (["a b", "c", "d e", "f"], [0, 0, 1, 1])),
    ],
)
def test_make_document_windows(docs, window_len, expected_windowed, has_y):
    kwargs = {}
    if has_y:
        y = [str(i) for i in range(len(docs))]
        kwargs = {"y": y}

    windowed, indices, windowed_y = make_document_windows(docs, window_len, **kwargs)

    assert (windowed, indices) == expected_windowed

    if has_y:
        assert [str(i) for i in indices] == windowed_y
    else:
        assert windowed_y is None


@pytest.mark.parametrize("output_cls", [PredictOutput, EmbedOutput])
@pytest.mark.parametrize(
    "unpooled_data,indices,pooling,pooled_data",
    [
        # One window, mean
        ([[0, 1]], [0], WindowPooling.MEAN, [[0, 1]]),
        # One window, max
        ([[0, 1]], [0], WindowPooling.MAX, [[0, 1]]),
        # One window, min
        ([[0, 1]], [0], WindowPooling.MIN, [[0, 1]]),
        # Two windows, same doc, mean
        ([[0, 1], [1, 0]], [0, 0], WindowPooling.MEAN, [[0.5, 0.5]]),
        # Two windows, same doc, max
        ([[0, 1], [1, 0]], [0, 0], WindowPooling.MAX, [[1, 1]]),
        # Two windows, same doc, min
        ([[0, 1], [1, 0]], [0, 0], WindowPooling.MIN, [[0, 0]]),
        # Two windows, different docs, mean
        ([[0, 1], [1, 0]], [0, 1], WindowPooling.MEAN, [[0, 1], [1, 0]]),
        # Two windows, different docs, max
        ([[0, 1], [1, 0]], [0, 1], WindowPooling.MAX, [[0, 1], [1, 0]]),
        # Two windows, different docs, min
        ([[0, 1], [1, 0]], [0, 1], WindowPooling.MIN, [[0, 1], [1, 0]]),
    ],
)
def test_pool_document_windows(
    output_cls, unpooled_data, indices, pooling, pooled_data
):
    unpooled_df = pd.DataFrame(unpooled_data).rename(str, axis=1)
    pooled_df = pd.DataFrame(pooled_data).rename(str, axis=1)

    if output_cls == PredictOutput:
        actual_output = PredictOutput(y_pred_proba=unpooled_df)
    elif output_cls == EmbedOutput:
        actual_output = EmbedOutput(X_embedded=unpooled_df)
    else:
        raise TypeError(
            f"Unknown class for document window pooling: '{output_cls.__name__}'"
        )

    pool_document_windows(actual_output, indices, pooling=pooling)

    if output_cls == PredictOutput:
        pd.testing.assert_frame_equal(actual_output.y_pred_proba, pooled_df)
    elif output_cls == EmbedOutput:
        for expected, actual in zip(np.array(pooled_data), actual_output.X_embedded):
            np.testing.assert_array_equal(expected, actual)


def test_pool_document_windows_validation():
    # Embedding output without pooling should throw an error
    embed_output = EmbedOutput(
        X_embedded=[np.array([[0, 1], [1, 0]])], embed_tokens=[["a", "b"]]
    )
    with pytest.raises(ValueError):
        pool_document_windows(embed_output, [])

    # Train output is unsupported
    train_output = TrainOutput(
        valid_loss=0.0, valid_accuracy=0.0, train_loss=0.0, labels=[]
    )
    with pytest.raises(TypeError):
        pool_document_windows(train_output, [])

    # Output and indices length must match
    embed_output = EmbedOutput(X_embedded=[np.array([0, 1]), np.array([1, 0])])

    for bad_indices_len in (0, 1, 3):
        with pytest.raises(ValueError):
            pool_document_windows(embed_output, list(range(bad_indices_len)))

    # Pooling value must be valid
    with pytest.raises(ValueError):
        pool_document_windows(embed_output, [0, 1], pooling="bad")



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

import gobbli.io
from gobbli.docker import maybe_mount, run_container
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import EmbedMixin, PredictMixin, TrainMixin
from gobbli.util import assert_type, escape_line_delimited_texts


class Transformer(BaseModel, TrainMixin, PredictMixin, EmbedMixin):
    """
    Classifier/embedding wrapper for any of the Transformers from
    `pytorch-transformers <https://github.com/huggingface/pytorch-transformers>`__.
    """

    _BUILD_PATH = Path(__file__).parent

    _TRAIN_INPUT_FILE = "train.tsv"
    _VALID_INPUT_FILE = "dev.tsv"
    _TEST_INPUT_FILE = "test.tsv"
    _LABELS_INPUT_FILE = "labels.tsv"
    _CONFIG_OVERRIDE_FILE = "config.json"

    _TRAIN_OUTPUT_CHECKPOINT = "checkpoint"
    _VALID_OUTPUT_FILE = "valid_results.json"
    _TEST_OUTPUT_FILE = "test_results.tsv"

    _EMBEDDING_INPUT_FILE = "input.tsv"
    _EMBEDDING_OUTPUT_FILE = "embeddings.jsonl"

    _CONTAINER_CACHE_DIR = Path("/cache")

    def init(self, params: Dict[str, Any]):
        """
        See :meth:`gobbli.model.base.BaseModel.init`.

        Transformer parameters:

        - ``transformer_model`` (:obj:`str`): Name of a transformer model architecture to use.
          For training/prediction, the value should be one such that
          ``from pytorch_transformers import <value>ForSequenceClassification`` is
          a valid import.  ex value = "Bert" ->
          ``from pytorch_transformers import BertForSequenceClassification``.  Note this means
          only a subset of the pytorch_transformers models are supported for these tasks -- search
          `the docs <https://huggingface.co/pytorch-transformers/search.html?q=forsequenceclassification&check_keywords=yes&area=default>`__ to see which ones you can use.
          For embedding generation, the import is ``<value>Model``, so any pytorch_transformer
          model is supported.
        - ``transformer_weights`` (:obj:`str`): Name of the pretrained weights to use.
          See the `pytorch-transformers docs <https://huggingface.co/pytorch-transformers/pretrained_models.html>`__
          for supported values.  These depend on the ``transformer_model`` chosen.
        - ``config_overrides`` (:obj:`dict`): Dictionary of keys and values that will
          override config for the model.
        - ``max_seq_length``: Truncate all sequences to this length after tokenization.
          Used to save memory.
        - ``lr``: Learning rate for the AdamW optimizer.
        - ``adam_eps``: Epsilon value for the AdamW optimizer.

        Note that gobbli relies on pytorch-transformers to perform validation on these parameters,
        so initialization errors may not be caught until model runtime.
        """
        self.transformer_model = "Bert"
        self.transformer_weights = "bert-base-uncased"
        self.config_overrides = {}  # type: Dict[str, Any]
        self.max_seq_length = 128
        self.lr = 5e-5
        self.adam_eps = 1e-8

        for name, value in params.items():
            if name == "transformer_model":
                self.transformer_model = value
            elif name == "transformer_weights":
                self.transformer_weights = value
            elif name == "config_overrides":
                assert_type(name, value, dict)
                self.config_overrides = value
            elif name == "max_seq_length":
                assert_type(name, value, int)
                self.max_seq_length = value
            elif name == "lr":
                assert_type(name, value, float)
                self.lr = value
            elif name == "adam_eps":
                assert_type(name, value, float)
                self.adam_eps = value
            else:
                raise ValueError(f"Unknown param '{name}'")

    @property
    def image_tag(self) -> str:
        """
        Returns:
          The Docker image tag to be used for the transformer container.
        """
        return "gobbli-transformer"

    def _build(self):
        self.docker_client.images.build(
            path=str(Transformer._BUILD_PATH),
            tag=self.image_tag,
            **self._base_docker_build_kwargs,
        )

    @staticmethod
    def _get_checkpoint(
        user_checkpoint: Optional[Path], context: ContainerTaskContext
    ) -> Tuple[Optional[Path], Optional[Path]]:
        """
        Determines the host checkpoint directory and container checkpoint directory
        using the user-requested checkpoint (if any) and the container context.

        Args:
          user_checkpoint: An optional checkpoint passed in by the user.  If the user doesn't
            pass one, use the default pretrained checkpoint.
          context: The container context to create the checkpoint in.

        Returns:
          A 2-tuple: the host checkpoint directory (if any) and
            the container checkpoint directory (if any)
        """
        if user_checkpoint is None:
            host_checkpoint_dir = None
            container_checkpoint_dir = None
        else:
            host_checkpoint_dir = user_checkpoint
            container_checkpoint_dir = context.container_root_dir / "checkpoint"

        return host_checkpoint_dir, container_checkpoint_dir

    def _get_weights(
        self, container_checkpoint_dir: Optional[Path]
    ) -> Union[str, Path]:
        """
        Determine the weights to pass to the run_model script.  If we don't have a
        checkpoint, we'll use the pretrained weights.  Otherwise, we should use the
        checkpoint weights.
        """
        if container_checkpoint_dir is None:
            return self.transformer_weights
        else:
            return container_checkpoint_dir

    @property
    def host_cache_dir(self):
        """
        Directory to be used for downloaded pytorch-transformers files.
        Should be the same across all instances of the class, since these are
        generally static model weights/config files that can be reused.
        """
        cache_dir = Transformer.model_class_dir() / "cache"
        cache_dir.mkdir(exist_ok=True, parents=True)
        return cache_dir

    def _write_input(self, X: List[str], labels: Optional[List[str]], input_path: Path):
        """
        Write the given input texts and (optionally) labels to the file pointed to by
        ``input_path``.
        """
        df = pd.DataFrame({"Text": X})

        if labels is not None:
            df["Label"] = labels

        df.to_csv(input_path, sep="\t", index=False)

    def _write_labels(self, labels: List[str], labels_path: Path):
        """
        Write the given labels to the file pointed at by ``labels_path``.
        """
        labels_path.write_text(escape_line_delimited_texts(labels))

    def _write_config(self, config_path: Path):
        """
        Write our model configuration overrides to the given path.
        """
        with open(config_path, "w") as f:
            json.dump(self.config_overrides, f)

    def _train(
        self, train_input: gobbli.io.TrainInput, context: ContainerTaskContext
    ) -> gobbli.io.TrainOutput:

        self._write_input(
            train_input.X_train,
            train_input.y_train,
            context.host_input_dir / Transformer._TRAIN_INPUT_FILE,
        )
        self._write_input(
            train_input.X_valid,
            train_input.y_valid,
            context.host_input_dir / Transformer._VALID_INPUT_FILE,
        )
        self._write_config(context.host_input_dir / Transformer._CONFIG_OVERRIDE_FILE)

        labels = train_input.labels()
        self._write_labels(
            labels, context.host_input_dir / Transformer._LABELS_INPUT_FILE
        )

        # Determine checkpoint to use
        host_checkpoint_dir, container_checkpoint_dir = self._get_checkpoint(
            train_input.checkpoint, context
        )

        cmd = (
            "python3 run_model.py"
            " train"
            f" --input-dir {context.container_input_dir}"
            f" --output-dir {context.container_output_dir}"
            f" --config-overrides {context.container_input_dir / Transformer._CONFIG_OVERRIDE_FILE}"
            f" --model {self.transformer_model}"
            f" --weights {self._get_weights(container_checkpoint_dir)}"
            f" --cache-dir {Transformer._CONTAINER_CACHE_DIR}"
            f" --max-seq-length {self.max_seq_length}"
            f" --train-batch-size {train_input.train_batch_size}"
            f" --valid-batch-size {train_input.valid_batch_size}"
            f" --num-train-epochs {train_input.num_train_epochs}"
            f" --lr {self.lr}"
            f" --adam-eps {self.adam_eps}"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the checkpoint in the container if needed
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint_dir, container_checkpoint_dir
        )

        # Mount the cache directory
        maybe_mount(
            run_kwargs["volumes"], self.host_cache_dir, Transformer._CONTAINER_CACHE_DIR
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        # Read in the generated evaluation results
        with open(context.host_output_dir / Transformer._VALID_OUTPUT_FILE, "r") as f:
            results = json.load(f)

        return gobbli.io.TrainOutput(
            valid_loss=results["mean_valid_loss"],
            valid_accuracy=results["valid_accuracy"],
            train_loss=results["mean_train_loss"],
            labels=labels,
            checkpoint=context.host_output_dir / Transformer._TRAIN_OUTPUT_CHECKPOINT,
            _console_output=container_logs,
        )

    def _read_predictions(self, predict_path: Path):
        return pd.read_csv(predict_path, sep="\t")

    def _predict(
        self, predict_input: gobbli.io.PredictInput, context: ContainerTaskContext
    ) -> gobbli.io.PredictOutput:

        self._write_input(
            predict_input.X, None, context.host_input_dir / Transformer._TEST_INPUT_FILE
        )
        self._write_config(context.host_input_dir / Transformer._CONFIG_OVERRIDE_FILE)

        labels = predict_input.labels
        self._write_labels(
            labels, context.host_input_dir / Transformer._LABELS_INPUT_FILE
        )

        host_checkpoint_dir, container_checkpoint_dir = self._get_checkpoint(
            predict_input.checkpoint, context
        )

        cmd = (
            "python3 run_model.py"
            " predict"
            f" --input-dir {context.container_input_dir}"
            f" --output-dir {context.container_output_dir}"
            f" --config-overrides {context.container_input_dir / Transformer._CONFIG_OVERRIDE_FILE}"
            f" --model {self.transformer_model}"
            f" --weights {self._get_weights(container_checkpoint_dir)}"
            f" --cache-dir {Transformer._CONTAINER_CACHE_DIR}"
            f" --max-seq-length {self.max_seq_length}"
            f" --predict-batch-size {predict_input.predict_batch_size}"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the checkpoint in the container if needed
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint_dir, container_checkpoint_dir
        )

        # Mount the cache directory
        maybe_mount(
            run_kwargs["volumes"], self.host_cache_dir, Transformer._CONTAINER_CACHE_DIR
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        return gobbli.io.PredictOutput(
            y_pred_proba=self._read_predictions(
                context.host_output_dir / Transformer._TEST_OUTPUT_FILE
            ),
            _console_output=container_logs,
        )

    def _read_embeddings(
        self, embed_path: Path, pooling: gobbli.io.EmbedPooling
    ) -> Tuple[List[np.ndarray], Optional[List[List[str]]]]:

        embeddings = []  # type: List[np.ndarray]
        doc_tokens = []  # type: List[List[str]]
        with open(embed_path, "r") as f:
            for line in f:
                line_json = json.loads(line)
                embeddings.append(np.array(line_json["embedding"]))
                if pooling == gobbli.io.EmbedPooling.NONE:
                    doc_tokens.append(line_json["tokens"])

        tokens = None
        if pooling == gobbli.io.EmbedPooling.NONE:
            tokens = doc_tokens

        return embeddings, tokens

    def _embed(
        self, embed_input: gobbli.io.EmbedInput, context: ContainerTaskContext
    ) -> gobbli.io.EmbedOutput:
        self._write_input(
            embed_input.X,
            None,
            context.host_input_dir / Transformer._EMBEDDING_INPUT_FILE,
        )
        self._write_config(context.host_input_dir / Transformer._CONFIG_OVERRIDE_FILE)

        host_checkpoint_dir, container_checkpoint_dir = self._get_checkpoint(
            embed_input.checkpoint, context
        )

        cmd = (
            "python3 run_model.py"
            " embed"
            f" --input-dir {context.container_input_dir}"
            f" --output-dir {context.container_output_dir}"
            f" --config-overrides {context.container_input_dir / Transformer._CONFIG_OVERRIDE_FILE}"
            f" --model {self.transformer_model}"
            f" --weights {self._get_weights(container_checkpoint_dir)}"
            f" --cache-dir {Transformer._CONTAINER_CACHE_DIR}"
            f" --max-seq-length {self.max_seq_length}"
            f" --embed-batch-size {embed_input.embed_batch_size}"
            f" --embed-pooling {embed_input.pooling.value}"
            f" --embed-layer -2"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the checkpoint in the container if needed
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint_dir, container_checkpoint_dir
        )

        # Mount the cache directory
        maybe_mount(
            run_kwargs["volumes"], self.host_cache_dir, Transformer._CONTAINER_CACHE_DIR
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        X_embedded, embed_tokens = self._read_embeddings(
            context.host_output_dir / Transformer._EMBEDDING_OUTPUT_FILE,
            embed_input.pooling,
        )

        return gobbli.io.EmbedOutput(
            X_embedded=X_embedded,
            embed_tokens=embed_tokens,
            _console_output=container_logs,
        )



from abc import abstractmethod
from pathlib import Path
from typing import List, Set, Tuple

from gobbli.dataset.base import BaseDataset


class NestedFileDataset(BaseDataset):
    """
    A dataset downloaded as an archive from some URL
    and composed of the following directory structure:

    <train_folder>/
      <label1>/
        data1
        data2
      <label2>/
        data1
        data2
    <test_folder>/
      <label1>/
        data1
        data2
      <label2>/
        data1
        data2
    """

    TRAIN_X_FILE = "train_X.csv"
    TRAIN_Y_FILE = "train_y.csv"

    TEST_X_FILE = "test_X.csv"
    TEST_Y_FILE = "test_Y.csv"

    # Null byte _shouldn't_ be embedded in any data files...
    DELIMITER = "\x00"

    @abstractmethod
    def labels(self) -> Set[str]:
        """
        Return the set of folder names that should be considered
        labels in each directory.
        """
        raise NotImplementedError

    @abstractmethod
    def download(self, data_dir: Path) -> Path:
        """
        Download and extract the dataset archive into the given data dir.
        Return the resulting path.
        """
        raise NotImplementedError

    @abstractmethod
    def folders(self) -> Tuple[Path, Path]:
        """
        Return relative paths to the train and test folders, respectively, from the
        top level of the extracted archive.
        """
        raise NotImplementedError

    @abstractmethod
    def read_source_file(self, file_path: Path) -> str:
        """
        Read the text from a source file.  Used to account for per-dataset encodings
        and other format differences.
        """
        raise NotImplementedError

    def _build(self):
        data_dir = self.data_dir()
        data_dir.mkdir(exist_ok=True, parents=True)

        self.download(data_dir)

        train_folder, test_folder = self.folders()

        self._load_folder(
            data_dir / train_folder,
            data_dir / self.TRAIN_X_FILE,
            data_dir / self.TRAIN_Y_FILE,
        )

        self._load_folder(
            data_dir / test_folder,
            data_dir / self.TEST_X_FILE,
            data_dir / self.TEST_Y_FILE,
        )

    def _load_folder(self, folder, X_file, y_file):
        """
        Combine a nested directory structure into a single output file.
        Assume the directory names are category names, and each file in
        the directory is a separate row assigned to that category.
        Write output to the given output file.
        """
        X = []
        y = []

        labels = self.labels()

        for category_dir in folder.iterdir():
            category_name = category_dir.name

            # Skip folders not named for labels
            if category_name not in labels:
                continue

            for data_file in category_dir.iterdir():
                X.append(self.read_source_file(data_file))
                y.append(category_name)

        X_file.write_text(NestedFileDataset.DELIMITER.join(X))
        y_file.write_text(NestedFileDataset.DELIMITER.join(y))

    def _read_data_file(self, filepath: Path) -> List[str]:
        return filepath.read_text().split(NestedFileDataset.DELIMITER)

    def _is_built(self) -> bool:
        data_files = (
            self.TRAIN_X_FILE,
            self.TRAIN_Y_FILE,
            self.TEST_X_FILE,
            self.TEST_Y_FILE,
        )
        return all((self.data_dir() / data_file).exists() for data_file in data_files)

    def X_train(self):
        return self._read_data_file(self.data_dir() / self.TRAIN_X_FILE)

    def y_train(self):
        return self._read_data_file(self.data_dir() / self.TRAIN_Y_FILE)

    def X_test(self):
        return self._read_data_file(self.data_dir() / self.TEST_X_FILE)

    def y_test(self):
        return self._read_data_file(self.data_dir() / self.TEST_Y_FILE)



import logging
import os
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Optional

import docker


def format_container_output(output: bytes) -> str:
    """
    Format the output of a Docker container for printing or logging.

    Args:
      output: Raw bytes output by the container.

    Returns:
      Output formatted as a string.
    """
    # Decode bytestring, remove trailing newlines that get inserted.
    return output.decode("utf-8").rstrip("\n")


def maybe_mount(
    volumes: Dict[str, Dict[str, Any]],
    host_path: Optional[Path],
    container_path: Optional[Path],
    mode: str = "rw",
):
    """
    If the given host path is not None, mutate the volumes to add an appropriate
    mapping for the path.  Otherwise, don't modify the volumes.

    Args:
      volumes: Dict of existing Docker volumes.
      host_path: An optional host path to be mounted.
      container_path: The path to be mounted if the host path exists.
      mode: Mount mode for the path.
    """
    if host_path is not None:
        volumes[str(host_path)] = {"bind": str(container_path), "mode": mode}


def run_container(
    client: docker.DockerClient,
    image_tag: str,
    cmd: str,
    logger: logging.Logger,
    **kwargs,
) -> str:
    """
    Run a container, appropriately display/handle output, and remove after
    finishing.  Resolve all host directories to be mounted as volumes in order to
    prevent errors related to docker mounting relative paths.

    Args:
      client: The Docker client to use to run the container.
      image_tag: Tag for the image to create the container from.
      cmd: Command to run in the container.
      logger: Logger which will be used to log debug information from the container.
      **kwargs: Additional arguments to pass to :meth:`docker.models.containers.ContainerCollection.run`.

    Returns:
      A string containing all output from the container.
    """
    # Resolve host paths
    run_kwargs = deepcopy(kwargs)
    volumes = run_kwargs.get("volumes", {})
    run_kwargs["volumes"] = {
        os.path.abspath(host_path): container_path
        for host_path, container_path in volumes.items()
    }

    logger.debug(f"Running container for image '{image_tag}' with command '{cmd}'")
    logger.debug(f"Container volumes: {run_kwargs['volumes']}")

    # Run as current UID to avoid files created by root
    container = client.containers.run(
        image_tag, cmd, user=os.geteuid(), group_add=[os.getegid()], **run_kwargs
    )

    try:
        for line in container.logs(stream=True):
            logger.debug(f"CONTAINER: {format_container_output(line)}")

        container_logs = format_container_output(container.logs())
        results = container.wait()
    finally:
        container.stop()
        container.remove()

    if results["StatusCode"] != 0:
        tail_logs = "\n".join(container_logs.split("\n")[-20:])
        err_msg = (
            f"Error running container (return code {results['StatusCode']})."
            " Last 20 lines of logs: \n"
            f"{tail_logs}"
        )
        if results["StatusCode"] == 137:
            err_msg += (
                "\n\nError code 137 indicates 'out of memory'.  Your input/model are likely too"
                " large to fit in RAM, although you should read the above error message to"
                " confirm.  Try adjusting parameters that affect memory usage, such as"
                " lowering the batch size or sequence length."
            )

        raise RuntimeError(err_msg)

    return container_logs



import logging
import random
from abc import ABC, abstractmethod
from pathlib import Path
from timeit import default_timer as timer
from typing import Any, List, Optional, Tuple

from sklearn.model_selection import train_test_split

import gobbli.io
from gobbli.util import format_duration, gobbli_dir

LOGGER = logging.getLogger(__name__)


def dataset_dir() -> Path:
    return gobbli_dir() / "dataset"


class BaseDataset(ABC):
    """
    Abstract base class for datasets used for benchmarking and testing.

    Derived classes should account for the following:
     - Dataset order should be consistent so limiting can work correctly
    """

    def __init__(self, *args, **kwargs):
        """
        Blank constructor needed to satisfy mypy
        """

    @classmethod
    def data_dir(cls) -> Path:
        return dataset_dir() / cls.__name__

    @classmethod
    def load(cls, *args, **kwargs) -> "BaseDataset":
        ds = cls(*args, **kwargs)

        if not ds._is_built():
            LOGGER.info("Dataset %s hasn't been built; building.", cls.__name__)
            start = timer()
            ds._build()
            end = timer()

            LOGGER.info(f"Dataset building finished in {format_duration(end - start)}.")

        return ds

    @abstractmethod
    def _is_built(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def _build(self):
        raise NotImplementedError

    @abstractmethod
    def X_train(self):
        raise NotImplementedError

    @abstractmethod
    def y_train(self):
        raise NotImplementedError

    @abstractmethod
    def X_test(self):
        raise NotImplementedError

    @abstractmethod
    def y_test(self):
        raise NotImplementedError

    def _get_train_valid(
        self, limit: Optional[int] = None, shuffle_seed: int = 1234
    ) -> Tuple[List[str], List[Any]]:
        """
        Return the X and y used for training and validation with the
        appropriate limit applied.  Shuffle first to minimize the possibility of
        getting only a single label in a small/limited dataset if it happens to be ordered
        by label.
        """
        X_train_valid = self.X_train()
        y_train_valid = self.y_train()

        # Shuffle the two simultaneously so text and label stay together
        X_y = list(zip(X_train_valid, y_train_valid))
        random.seed(shuffle_seed)
        random.shuffle(X_y)

        X_train_valid[:], y_train_valid[:] = zip(*X_y)

        if limit is not None:
            X_train_valid = X_train_valid[:limit]
            y_train_valid = y_train_valid[:limit]

        return X_train_valid, y_train_valid

    def train_input(
        self,
        train_batch_size: int = 32,
        valid_batch_size: int = 8,
        num_train_epochs: int = 3,
        valid_proportion: float = 0.2,
        split_seed: int = 1234,
        shuffle_seed: int = 1234,
        limit: Optional[int] = None,
    ) -> gobbli.io.TrainInput:
        if not self._is_built():
            raise ValueError("Dataset must be built before accessing train_input.")

        X_train_valid, y_train_valid = self._get_train_valid(
            limit=limit, shuffle_seed=shuffle_seed
        )

        X_train, X_valid, y_train, y_valid = train_test_split(
            X_train_valid,
            y_train_valid,
            test_size=valid_proportion,
            random_state=split_seed,
        )

        return gobbli.io.TrainInput(
            X_train=X_train,
            X_valid=X_valid,
            y_train=y_train,
            y_valid=y_valid,
            train_batch_size=train_batch_size,
            valid_batch_size=valid_batch_size,
            num_train_epochs=num_train_epochs,
        )

    def embed_input(
        self,
        embed_batch_size: int = 32,
        pooling: gobbli.io.EmbedPooling = gobbli.io.EmbedPooling.MEAN,
        limit: Optional[int] = None,
    ) -> gobbli.io.EmbedInput:
        if not self._is_built():
            raise ValueError("Dataset must be built before accessing embed_input.")

        X_test = self.X_test()
        if limit is not None:
            X_test = X_test[:limit]

        return gobbli.io.EmbedInput(
            X=X_test, embed_batch_size=embed_batch_size, pooling=pooling
        )

    def predict_input(
        self, predict_batch_size: int = 32, limit: Optional[int] = None
    ) -> gobbli.io.PredictInput:
        if not self._is_built():
            raise ValueError("Dataset must be built before accessing predict_input.")

        _, y_train_valid = self._get_train_valid(limit=limit)

        X_test = self.X_test()
        if limit is not None:
            X_test = X_test[:limit]

        return gobbli.io.PredictInput(
            X=X_test,
            predict_batch_size=predict_batch_size,
            labels=sorted(list(set(y_train_valid))),
        )



from gobbli.dataset.imdb import IMDBDataset
from gobbli.dataset.newsgroups import NewsgroupsDataset
from gobbli.dataset.trivial import TrivialDataset

__all__ = ["TrivialDataset", "NewsgroupsDataset", "IMDBDataset"]



from gobbli.model.bert import BERT
from gobbli.model.fasttext import FastText
from gobbli.model.majority import MajorityClassifier
from gobbli.model.mtdnn import MTDNN
from gobbli.model.random import RandomEmbedder
from gobbli.model.transformer import Transformer
from gobbli.model.use import USE

__all__ = [
    "BERT",
    "FastText",
    "MajorityClassifier",
    "MTDNN",
    "RandomEmbedder",
    "Transformer",
    "USE",
]
